from gaussian_splatting import GaussianModel, CameraTrainableGaussianModel
from gaussian_splatting.trainer import DepthTrainerWrapper, CameraTrainerWrapper
from lapisgs.dataset import RescaleTrainableCameraDataset
from .densifier import PartialDensificationTrainer
from .opacity_reset import PartialOpacityResetTrainerWrapper


def LapisTrainer(model: GaussianModel, scene_extent: float, *args, opacity_lr=0.05, **kwargs):
    return PartialOpacityResetTrainerWrapper(PartialDensificationTrainer, model, scene_extent, *args, opacity_lr=opacity_lr, **kwargs)


def DepthLapisTrainer(model: GaussianModel, scene_extent: float, *args, **kwargs):
    return DepthTrainerWrapper(LapisTrainer, model, scene_extent, *args, **kwargs)


def LapisCameraTrainer(model: CameraTrainableGaussianModel, scene_extent: float, dataset: RescaleTrainableCameraDataset, *args, **kwargs):
    return CameraTrainerWrapper(
        lambda model, scene_extent, dataset, *args, **kwargs: LapisTrainer(model, scene_extent, *args, **kwargs),
        model, scene_extent, dataset, *args, **kwargs
    )


def DepthLapisCameraTrainer(model: CameraTrainableGaussianModel, scene_extent: float, dataset: RescaleTrainableCameraDataset, *args, **kwargs):
    return CameraTrainerWrapper(
        lambda model, scene_extent, dataset, *args, **kwargs: DepthLapisTrainer(model, scene_extent, *args, **kwargs),
        model, scene_extent, dataset, *args, **kwargs
    )


# Aliases for default trainers
Trainer = LapisTrainer
CameraTrainer = LapisCameraTrainer
DepthTrainer = DepthLapisTrainer
DepthCameraTrainer = DepthLapisCameraTrainer



from typing import List
from gaussian_splatting import GaussianModel, Camera
from gaussian_splatting.dataset import CameraDataset, TrainableCameraDataset
from gaussian_splatting.trainer import CameraTrainerWrapper, NoopDensifier, DepthTrainerWrapper
from reduced_3dgs.shculling import VariableSHGaussianModel, SHCullingTrainerWrapper
from reduced_3dgs.combinations import CameraTrainableVariableSHGaussianModel
from lapisgs.trainer import LapisTrainer, PartialOpacityResetTrainerWrapper
from .trainer import PartialPrunerInDensifyTrainerWrapper
from .importance import ImportancePruner


def PartialFullPrunerInDensifyTrainer(
        model: GaussianModel,
        scene_extent: float,
        dataset: List[Camera],
        *args,
        importance_prune_from_iter=15000,
        importance_prune_until_iter=20000,
        importance_prune_interval: int = 1000,
        importance_score_resize=None,
        importance_prune_type="comprehensive",
        importance_prune_percent=0.1,
        importance_prune_thr_important_score=None,
        importance_prune_thr_v_important_score=3.0,
        importance_prune_thr_max_v_important_score=None,
        importance_prune_thr_count=1,
        importance_prune_thr_T_alpha=1.0,
        importance_prune_thr_T_alpha_avg=0.001,
        importance_v_pow=0.1,
        **kwargs):
    return PartialPrunerInDensifyTrainerWrapper(
        lambda model, scene_extent, dataset: ImportancePruner(
            NoopDensifier(model),
            dataset,
            importance_prune_from_iter=importance_prune_from_iter,
            importance_prune_until_iter=importance_prune_until_iter,
            importance_prune_interval=importance_prune_interval,
            importance_score_resize=importance_score_resize,
            importance_prune_type=importance_prune_type,
            importance_prune_percent=importance_prune_percent,
            importance_prune_thr_important_score=importance_prune_thr_important_score,
            importance_prune_thr_v_important_score=importance_prune_thr_v_important_score,
            importance_prune_thr_max_v_important_score=importance_prune_thr_max_v_important_score,
            importance_prune_thr_count=importance_prune_thr_count,
            importance_prune_thr_T_alpha=importance_prune_thr_T_alpha,
            importance_prune_thr_T_alpha_avg=importance_prune_thr_T_alpha_avg,
            importance_v_pow=importance_v_pow,
        ),
        model, scene_extent, dataset,
        *args, **kwargs
    )


def DepthPartialFullPrunerInDensifyTrainer(model: GaussianModel, scene_extent: float, dataset: TrainableCameraDataset, *args, **kwargs):
    return DepthTrainerWrapper(PartialFullPrunerInDensifyTrainer, model, scene_extent, *args, dataset=dataset, **kwargs)


def PartialOpacityResetPrunerInDensifyTrainer(
        model: GaussianModel,
        scene_extent: float,
        dataset: CameraDataset,
        *args, **kwargs):
    return PartialOpacityResetTrainerWrapper(
        lambda model, scene_extent, *args, **kwargs: DepthPartialFullPrunerInDensifyTrainer(model, scene_extent, dataset, *args, **kwargs),
        model, scene_extent,
        *args, **kwargs
    )


LapisFullTrainer = PartialOpacityResetPrunerInDensifyTrainer


def SHCullingLapisTrainer(
    model: VariableSHGaussianModel,
        scene_extent: float,
        dataset: CameraDataset,
        *args, **kwargs):
    return SHCullingTrainerWrapper(
        lambda model, scene_extent, dataset, *args, **kwargs: LapisTrainer(model, scene_extent, *args, **kwargs),
        model, scene_extent, dataset,
        *args, **kwargs
    )


def SHCullingLapisFullTrainer(
    model: VariableSHGaussianModel,
        scene_extent: float,
        dataset: CameraDataset,
        *args, **kwargs):
    return SHCullingTrainerWrapper(
        PartialOpacityResetPrunerInDensifyTrainer,
        model, scene_extent, dataset,
        *args, **kwargs
    )


def CameraLapisFullTrainer(
        model: CameraTrainableVariableSHGaussianModel,
        scene_extent: float,
        dataset: TrainableCameraDataset,
        *args, **kwargs):
    return CameraTrainerWrapper(
        PartialOpacityResetPrunerInDensifyTrainer,
        model, scene_extent, dataset,
        *args, **kwargs
    )


def CameraSHCullingLapisTrainer(
        model: CameraTrainableVariableSHGaussianModel,
        scene_extent: float,
        dataset: TrainableCameraDataset,
        *args, **kwargs):
    return CameraTrainerWrapper(
        SHCullingLapisTrainer,
        model, scene_extent, dataset,
        *args, **kwargs
    )


def CameraSHCullingLapisFullTrainer(
        model: CameraTrainableVariableSHGaussianModel,
        scene_extent: float,
        dataset: TrainableCameraDataset,
        *args, **kwargs):
    return CameraTrainerWrapper(
        SHCullingLapisFullTrainer,
        model, scene_extent, dataset,
        *args, **kwargs
    )



from typing import Callable
import torch

from gaussian_splatting import GaussianModel
from gaussian_splatting.trainer import AbstractTrainer
from gaussian_splatting.trainer.opacity_reset import OpacityResetter, replace_tensor_to_optimizer


class PartialOpacityResetter(OpacityResetter):
    def __init__(
            self, base_trainer: AbstractTrainer,
            *args, **kwargs
    ):
        super().__init__(base_trainer, *args, **kwargs)
        self.size_fixed_gs = base_trainer.model.get_opacity.shape[0]

    def optim_step(self):
        with torch.no_grad():
            if self.opacity_reset_from_iter <= self.curr_step <= self.opacity_reset_until_iter and self.curr_step % self.opacity_reset_interval == 0:
                # https://github.com/nus-vv-streams/lapis-gs/blob/12dcda37ed43838d7407b28675bc26b7364ae431/scene/gaussian_model.py#L226
                opacities_new_dynamic = self.model.inverse_opacity_activation(torch.min(self.model.get_opacity[self.size_fixed_gs:], torch.ones_like(self.model.get_opacity[self.size_fixed_gs:])*0.01))
                opacities_new = torch.cat((self.model._opacity[:self.size_fixed_gs], opacities_new_dynamic), dim=0)
                optimizable_tensors = replace_tensor_to_optimizer(self.optimizer, opacities_new, "opacity")
                self.model._opacity = optimizable_tensors["opacity"]
                torch.cuda.empty_cache()
        return super().optim_step()


def PartialOpacityResetTrainerWrapper(
        base_trainer_constructor: Callable[..., AbstractTrainer],
        model: GaussianModel,
        scene_extent: float,
        *args,
        opacity_reset_from_iter=3000,
        opacity_reset_until_iter=15000,
        opacity_reset_interval=3000,
        **kwargs) -> PartialOpacityResetter:
    return PartialOpacityResetter(
        base_trainer=base_trainer_constructor(model, scene_extent, *args, **kwargs),
        opacity_reset_from_iter=opacity_reset_from_iter,
        opacity_reset_until_iter=opacity_reset_until_iter,
        opacity_reset_interval=opacity_reset_interval,
    )



#
# Copyright (C) 2023, Inria
# GRAPHDECO research group, https://team.inria.fr/graphdeco
# All rights reserved.
#
# This software is free for non-commercial, research and evaluation use
# under the terms of the LICENSE.md file.
#
# For inquiries contact  george.drettakis@inria.fr
#

from setuptools import setup, find_packages

with open("README.md", "r", encoding='utf8') as fh:
    long_description = fh.read()

packages = ['lapisgs'] + ["lapisgs." + package for package in find_packages(where="lapisgs")]
packages += ['lapisgs.trainer.extensions.' + package for package in find_packages(where="lapisgs/trainer/extensions")]

setup(
    name="lapisgs",
    version='1.0.0',
    author='yindaheng98',
    author_email='yindaheng98@gmail.com',
    url='https://github.com/yindaheng98/lapis-gs',
    description=u'Refactored python training and inference code for 3D Gaussian Splatting',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    packages=packages,
    package_dir={
        'lapisgs': 'lapisgs',
    },
    install_requires=[
        'gaussian-splatting',
        'reduced-3dgs'
    ]
)



import torch
import json

from gaussian_splatting.dataset import JSONCameraDataset, TrainableCameraDataset
from gaussian_splatting.camera import camera2dict, dict2camera, Camera
from gaussian_splatting.utils import quaternion_to_matrix


class RescaleCameraDatasetIface:
    def save_cameras(self, path):
        # https://github.com/yindaheng98/gaussian-splatting/blob/56576b647d9c5bd05300f5640cd03a8c75a760bc/gaussian_splatting/dataset/dataset.py#L24
        cameras = []
        for id, camera in enumerate(self):
            cameradict = camera2dict(camera, id)
            fullimage_width, fullimage_height = camera.image_width, camera.image_height
            if "fullimage_width" in camera.custom_data and "fullimage_height" in camera.custom_data:
                fullimage_width = camera.custom_data["fullimage_width"]
                fullimage_height = camera.custom_data["fullimage_height"]
            cameradict["fullimage_width"] = fullimage_width
            cameradict["fullimage_height"] = fullimage_height
            cameras.append(cameradict)
        with open(path, 'w') as f:
            json.dump(cameras, f, indent=2)


class RescaleJSONCameraDataset(RescaleCameraDatasetIface, JSONCameraDataset):
    def __init__(self, path, load_depth=False, rescale_factor=1.0):
        # https://github.com/yindaheng98/gaussian-splatting/blob/56576b647d9c5bd05300f5640cd03a8c75a760bc/gaussian_splatting/dataset/dataset.py#L38
        with open(path, 'r') as f:
            self.json_cameras = json.load(f)
        self.load_depth = load_depth
        for camera in self.json_cameras:
            camera['fx'] = camera['fx'] / camera['width'] * camera['fullimage_width'] * rescale_factor
            camera['fy'] = camera['fy'] / camera['height'] * camera['fullimage_height'] * rescale_factor
            camera['width'] = round(camera['fullimage_width'] * rescale_factor)
            camera['height'] = round(camera['fullimage_height'] * rescale_factor)
        self.cameras = [dict2camera(
            camera, load_depth=self.load_depth,
            custom_data=dict(fullimage_width=camera['fullimage_width'], fullimage_height=camera['fullimage_height'])
        ) for camera in self.json_cameras]


class RescaleTrainableCameraDataset(TrainableCameraDataset):
    def save_cameras(self, path):
        # https://github.com/yindaheng98/gaussian-splatting/blob/56576b647d9c5bd05300f5640cd03a8c75a760bc/gaussian_splatting/dataset/camera_trainable.py#L68
        cameras = []
        for idx, camera in enumerate(self):
            cameras.append({
                **camera2dict(Camera(**{
                    **camera._asdict(),
                    'R': quaternion_to_matrix(self.quaternions[idx, ...]),
                    'T': self.Ts[idx, ...],
                }), idx),
                "exposure": self.exposures[idx, ...].detach().tolist(),
                "fullimage_width": camera.custom_data["fullimage_width"],
                "fullimage_height": camera.custom_data["fullimage_height"],
            })
        with open(path, 'w') as f:
            json.dump(cameras, f, indent=2)

    @classmethod
    def from_json(cls, path, load_depth=False, rescale_factor=1.0):
        # https://github.com/yindaheng98/gaussian-splatting/blob/56576b647d9c5bd05300f5640cd03a8c75a760bc/gaussian_splatting/dataset/camera_trainable.py#L83
        cameras = RescaleJSONCameraDataset(path, load_depth=load_depth, rescale_factor=rescale_factor)
        exposures = [(torch.tensor(camera['exposure'], dtype=torch.float) if 'exposure' in camera else torch.eye(3, 4)) for camera in cameras.json_cameras]
        return cls(cameras, exposures)
